/* hostio.c -
 * Copyright (C) 1987-2003 J.G.Harston, David Daniels

 See the accompanying file LICENSE for authors and terms of use.

   28-Jan-1993 JGH: This is the re-written interface to UNIX
            'cos I went and copied the Arc version on top of it!
   14-Apr-1993 JGH: Started converting for Mark2 C Version
   20-May-1993 JGH: Started ANSI-ising it
   21-May-1993 JGH: removed look_esc(), add addr= to OSWORD,
   tidied up MOS_FF0D (SWI)
   28-May-1993 JGH: Still problems getting ioctl() to work.  Had
               to add system("stty...") commands to startup() and finish()
   28-May-1993 JGH: Added OSFILE/PUT/GET for bbc
   31-May-1993 JGH: MOS_QUIT moved to z80.c, PC made a pointer
               OSGBPB for __bbc written, but where is mem[HLreg+0] returned?
   26-Jun-1993 JGH: *Basic load Z80Tube$Basic; *com checks Z80$Flag
   12-Jun-1995 JGH: Note: could do following
               on startup, read ioctl, do "stty -cooked", read ioctl
               then use ioctl to change, and on shutdown, set back
   30-Aug-2003 JGH: Now called hostio.c
*/

/* ==================================================================== */
/* RISC OS kernel interface						*/
/* ==================================================================== */
#ifdef Z80_IO_RO
#include <kernel.h>
#include <swis.h>
_kernel_oserror oserr;			/* Returned error information	*/
_kernel_oserror *oserr2;		/* Error from OSWORD 0		*/
_kernel_osfile_block ctrl;		/* OSFILE control block		*/
_kernel_osgbpb_block gbpb;		/* OSGBPB control block		*/
_kernel_swi_regs regs;			/* Used by OSWORD 0		*/
#endif

/* ==================================================================== */
/* UNIX kernel interface						*/
/* ==================================================================== */
#ifdef Z80_IO_UNIX
#include <time.h>
#include <sys/ioctl.h>
struct tchars z80chars;
struct sgttyb z80params;
#endif

/* ==================================================================== */
/* Windows kernel interface						*/
/* ==================================================================== */
#ifdef Z80_IO_WIN
#include <time.h>
#include <windows.h>
/* #include <winbase.h> */
#endif

/* ==================================================================== */
/* DOS kernel interface							*/
/* ==================================================================== */
#ifdef Z80_IO_DOS
#include <time.h>
#endif

/* ==================================================================== */
/* Escape handling							*/
/* Needs tidying up							*/
/* ==================================================================== */
#include <signal.h>

void esc_pressed(int signal)		/* Escape key pressed		*/
{
signal=signal;				/* Prevent optiser removal	*/
mem[0xFF80]=128;			/* Set z80's escape happened	*/
esc_claim();				/* Reclaim signal		*/
}

esc_claim()			/* Claim SIGINT (escape) signal	*/
{
signal(SIGINT, esc_pressed);
}

void esc_release(void)			/* Release SIGINT (escape)	*/
{
signal(SIGINT, SIG_DFL);
}

void esc_on(void)			/* Turn SIGINT on for CHR$27 */
{
#ifdef Z80_IO_UNIX
ioctl(stdin,TIOCGETC,&z80chars);
z80chars.t_intrc=27;
z80chars.t_quitc=-1;
ioctl(stdin,TIOCSETC,&z80chars);
#endif
}

void esc_off(void)			/* Turn SIGINT off */
{
#ifdef Z80_IO_UNIX
ioctl(stdin,TIOCGETC,&z80chars);
z80chars.t_intrc=-1;
z80chars.t_quitc=-1;
ioctl(stdin,TIOCSETC,&z80chars);
#endif
}

/* ==================================================================== */
/* Text output handling							*/
/* Needs work								*/
/* ==================================================================== */

/* Set output to 'semi-cooked' */
void set_cbreak(void)
{
#ifdef Z80_IO_UNIX
/*
system("stty cbreak -echo");  */     /* This REALLY should use ioctl() */

ioctl(stdout,TIOCGETP, &z80params);
z80params.sg_flags=((z80params.sg_flags & !RAW & !ECHO) | CBREAK);
ioctl(stdout,TIOCSETP, &z80params);
ioctl(stdin, TIOCGETP, &z80params);
z80params.sg_flags=((z80params.sg_flags & !RAW & !ECHO) | CBREAK);
ioctl(stdin, TIOCSETP, &z80params);
#endif
}

/* Set output to cooked */
void set_cooked(void)
{
#ifdef Z80_IO_UNIX
/*
system("stty -cbreak echo");  */  /* This also REALLY should use ioctl() */

ioctl(stdout,TIOCGETP, &z80params);
z80params.sg_flags=((z80params.sg_flags & !RAW & !CBREAK) | ECHO);
ioctl(stdout,TIOCSETP, &z80params);
ioctl(stdin, TIOCGETP, &z80params);
z80params.sg_flags=((z80params.sg_flags & !RAW & !CBREAK) | ECHO);
ioctl(stdin, TIOCSETP, &z80params);
#endif
}

/* ==================================================================== */
/* Various bits								*/
/* ==================================================================== */
int get_hex(char c) {
  if (c >= '0' && c <= '9') return c - '0';
  c = tolower(c);
  if (c >= 'a' && c <= 'f') return c - 'a' + 10;
  return 0;
}

/*
 * get_filetype - Returns the RISC OS file type of the named
 * file or -1 if file cannot be found or the name passed is
 * not that of a file
 */
int get_filetype(char *name) {
  _kernel_swi_regs regs;
  _kernel_oserror *error;
  regs.r[0] = 23;		/* Use OS_File call 23 to read file details */
  regs.r[1] = (int) name;
  error = _kernel_swi(OS_File, &regs, &regs);

/* R0 contains 1 if object is a file */

  if (error != NULL || regs.r[0] != 1) return -1;
  return regs.r[6];		/* R6 = file's filetype */
}


/* ==================================================================== */
/* Startup and finalisation						*/
/* ==================================================================== */
void io_start(void)
{
esc_claim();				/* Trap escapes			*/
esc_on();				/* Escape=CHR$27		*/
/* Need to set to semi-cooked */
disk_init();				/* Mount default images		*/
}

void io_finish(void)
{
esc_off();				/* Escape=?			*/
esc_release();				/* Escape=CHR$3, release Escape */
/* Need to set to cooked */
disk_dismount(-1);			/* Close any open disk images	*/
free(mem);				/* Release Z80 memory		*/
free(iomem);				/* Release I/O memory		*/
}


/* ==================================================================== */
/* Opcodes ED00 to ED0F call host MOS routines.				*/
/* ==================================================================== */

/* ==================================================================== */
/* OSQUIT - Quit the system						*/
/* ==================================================================== */
void MOS_QUIT(void)
{
io_finish();				/* Disconnect any environment	*/
exit(0);				/* And exit			*/
}


/* ==================================================================== */
/* OSCLI - Execute an command						*/
/* ==================================================================== */
void MOS_CLI(void)
{
int type;				/* filetype			*/

HLreg=Lreg | (Hreg << 8);		/* Point to string at HL	*/
addr=0;
for (addr=0; mem[HLreg]>31 && addr<250; HLreg++) iobuffer[addr++]=mem[HLreg];
iobuffer[addr]=0;			/* Copy HL string to iobuffer	*/
if (cli()) return;			/* Internal command		*/

#ifdef Z80_IO_RO
/* Check for a CPM-type command (filetype &2xx) */
  type = get_filetype(&iobuffer[tmp]);
  if (type >= 0x200 || type < 0x300) {
    addr = (type & 0xFF) << 8;
    if (load(&iobuffer[tmp], addr) != -1) {
      PC = &mem[0] + addr; 
      return;
    }
  }
_kernel_oscli("Unset Z80$Flag");
if(chk_err(_kernel_system(&iobuffer[tmp],0)))
	return;				/* Error occured, return	*/
_kernel_getenv("Z80$Flag", iobuffer, 250);
if(iobuffer[0] == 0) return;		/* Command completed externally */
addr=(get_hex(iobuffer[0]) << 12) | (get_hex(iobuffer[1]) << 8);
					/* Get address of file to load	*/
if(load(iobuffer+5,addr) != -1)
	PC=&mem[0]+addr;		/* Execute loaded file		*/
#endif

#ifndef Z80_IO_RO
system(&iobuffer[0]);			/* Needs to check for errors	*/
#endif
}


/* ==================================================================== */
/* OSBYTE - Byte commands						*/
/* ==================================================================== */
void MOS_BYTE(void)
{
if (debug & 2) printf("OsByte %02X,%02X%02X\n",Areg,Lreg,Hreg);

tmp=0;
switch (Areg) {
#ifndef Z80_IO_RO
  case 0:				/* Read host type		*/
    Lreg=Z80_HOST;
    return;
#endif

  case 0x7C:
  case 0x7E:				/* Clear Escape			*/
    mem[0xFF80]=0;			/* then drop though to tell OS	*/
    break;

  case 0x7D:				/* Set Escape			*/
    mem[0xFF80]=128;			/* then drop though to tell OS	*/
    break;

#ifndef Z80_IO_RO
  case 0x7F:				/* Read EOF#handle		*/
    Lreg=file_eof(Lreg);		/* Call file.c			*/
    return;

  case 0x81:				/* INKEY			*/
    if (Hreg == 0xFE) {			/* INKEY-ve			*/
      Lreg=(con_keyscan(Lreg+256) ? 0xFF : 0x00);
      Hreg=Lreg;
      return;
      }
    if (Hreg == 0xFF) {			/* INKEY-ve			*/
      if (Lreg) {
        Lreg=(con_keyscan(Lreg) ? 0xFF : 0x00);
        Hreg=Lreg;
        } else {
        Lreg=Z80_OS; Hreg=0;
        }
      return;
      }					/* INKEY+ve			*/
  case 1:				/* Read timed			*/
    tmp=clock()+(Lreg+((Hreg & 0x7f)<<8))*(CLOCKS_PER_SEC/100);
    for(;;) {
      if(kbhit() || (clock()>tmp)) break;
      }
    if(kbhit()) tmp=con_rdch(); else tmp=0xffff;
    Hreg=((Hreg & 0x80) ? tmp >> 8 : 0); /* &8000+n -> 16bit, n -> 8bit */
    if (tmp == 0xffff) Hreg=0xff;	 /* -1 -> 0xffff		*/
    if ((Lreg=tmp) == 27) mem[0xFF80]=128; /* Should check EscOn	*/
    Freg=(Freg & 254) | ((mem[0xFF80] & 128) >> 7); /* Escape	*/
    return;
#endif
  
  case 0x82:				/* Read Hi Order Address	*/
    Lreg=0; Hreg=0;
    return;

  case 0x83:				/* Read lowest PAGE		*/
    Lreg=0; Hreg=0x01;
    return;

  case 0x84:				/* Read HIMEM			*/
    Lreg=mem[6]; Hreg=mem[7];
    return;
  }

#ifdef Z80_IO_RO
chk_err(tmp=_kernel_osbyte(Areg,Lreg,Hreg));
#endif

Lreg=tmp & 255;				/* Extract registers from tmp	*/
Hreg=(tmp & 0xFF00) >> 8;
Freg=(Freg & 254) | ((tmp & 65536) >> 16);
}


/* ==================================================================== */
/* OSSWI - Do an SWI call - this has be here as MOS_WORD calls it	*/
/* ==================================================================== */

/* Format of control block:
 * HL=>0    number of in bytes, (number of regs)*4+8, usually 40
 *     1    number of out bytes, (number of regs)*4+8, usually 40
 *     2    outward translate flags, b0=r0, b1=r1, etc for Z80->Arm
 *     3    inward translate flags, b0=r0, b1=r1, etc for Arm-Z80
 *     4-7  SWI number
 *     8-11 r0
 *    12-15 r1
 *     etc
 *    36-39 r7
 */
void MOS_FF0D()
{
}


/* ==================================================================== */
/* Do an OSWORD								*/
/* Some calls need addresses translated:				*/
/*                             0123456789ABCDEF
   0: Read a line              01
   1: Read TIME
   2: Write TIME
   3: Read intervel timer
   4: Write interval timer
   5: Read memory
   6: Write memory
   7: SOUND
   8: ENVELOPE
   9: Read POINT(x,y)
  10: Read character definition
  11: Read palette
  12: Write palette
  13: Read graphics cursors
  14: Read TIME$
  15: Write TIME$
  16: Econet Transmit              456789AB
  17: Econet Receive                56789ABC
  18: Read RPC parameters
  19: Read/Set station info
  20: NFS Fileserver functions
  21: Mouse                    0     6789
  22: Screen address
  90: HADFS Read sectors       6 2345
 127: Read sectors              1234
 190: Disassemble code
 191: Assemble code
 200: SWI interface     (&ED0D)
 255: I/O processor memory access
*/
/* ==================================================================== */
void MOS_WORD(void)
{
int c;
int l;

HLreg=Lreg | (Hreg << 8);
addr32=mem[HLreg] | (mem[HLreg+1] << 8) | (mem[HLreg+2] << 16) | (mem[HLreg+3] << 24);
addr=addr32 & 0xFFFF;
if (debug & 2) printf("OsWord %02X,%04X\n",Areg,HLreg);

switch (Areg) {
  case 0:				/* Read a line of input		*/
#ifdef Z80_IO_RO
    regs.r[0]=addr+(int)&mem[0];	/* ARM address of input iobuffer */
    regs.r[1]=mem[HLreg+2];		/* iobuffer size		*/
    regs.r[2]=mem[HLreg+3];		/* min. ASCII			*/
    regs.r[3]=mem[HLreg+4];		/* max. ASCII			*/
    oserr2=_kernel_swi_c(0x0E, &regs, &regs, &tmp);
    Freg=(tmp & 1);
    Hreg=regs.r[1];			/* Returned length		*/
    if (oserr2 != NULL) z80error(oserr2->errnum, oserr2->errmess, "");
#else
    tmp=con_readln(addr, mem[HLreg+2], mem[HLreg+3], mem[HLreg+4]);
    if (tmp < 0) { Freg=((Freg & 254) | 1); }
    else { Hreg=tmp; }
#endif
    return;
    break;

#ifndef Z80_IO_RO
  case 1:				/* Read TIME			*/
    tmp=(int)clock()/(CLOCKS_PER_SEC/100);    
    mem[HLreg]=tmp;
    mem[HLreg+1]=tmp >> 8;
    mem[HLreg+2]=tmp >> 16;
    mem[HLreg+3]=tmp >> 24;
    mem[HLreg+4]=0;
    return;
    break;
#endif

  case 5:				/* Read from memory		*/
    mem[HLreg+4]=*(char *)addr32;
    return;
    break;

  case 6:				/* Write to memory		*/
    *(char *)addr32=mem[HLreg+4];
    return;
    break;

  case 127:				/* FM direct disk I/O		*/
    osword_127();
    return;

#ifdef XxXx
  case 190:				/* Disassembly routine		*/
    if (mem[HLreg+2] == 80) {		/* Z80 disassem routine		*/
      mem[HLreg+3]=1; mem[HLreg+4]='?'; mem[HLreg+5]=13;
      return;
    }
    break;

  case 191:				/* Assembly routine		*/
    return;
    break;
#endif

  case 200:				/* SWI Interface		*/
    MOS_FF0D();				/* Pass on to ED0D handler	*/
    return;
    break;

  case 255:				/* I/O processor memory access */
    osword_255();
    return;

  default:				/* All other calls		*/
#ifdef Z80_IO_RO
/* NB: Some calls need to have addresses relocated */
    for(l=0; l<250 & HLreg+l < 0xFFFF; l++) iobuffer[l]=mem[HLreg+l];
    chk_err(_kernel_osword(Areg, (int *)iobuffer));
    for(l=0; l<250 & HLreg+l < 0xFFFF; l++) mem[HLreg+l]=iobuffer[l];
#endif
    return;
    break;
  }
}


/* ==================================================================== */
/* OSWRCH - Send a character to output stream				*/
/* ==================================================================== */
void MOS_WRCH(void)
{
#ifdef Z80_IO_RO
chk_err(_kernel_oswrch(Areg));
#else
con_wrch(Areg);
#endif
}


/* ==================================================================== */
/* OSRDCH - Request an character from input stream			*/
/* ==================================================================== */
void MOS_RDCH(void)
{
#ifdef Z80_IO_RO
chk_err(tmp=_kernel_osrdch());
#endif
/*
If not RISC OS:
Escape trapped and causes FF80 to be set
Escape=CHR$27
Does Escape cause getch() to return?
*/
#ifndef Z80_IO_RO
tmp=(con_rdch() & 0xFF);
#endif
Areg=tmp;
if (Areg == 27) mem[0xFF80]=128;
Freg=(Freg & 254) | ((mem[0xFF80] & 128) >> 7);		/* Escape	*/
}


/* Adjust address of &0000xxxx to point to real memory */
void adjustadd(int *n, int *m)
{
#ifdef Z80_IO_RO
if ((*n & 0xFFFF0000) == 0) {
  *n=*n+(int)&mem[0];
  if (m != 0) *m=*m+(int)&mem[0];
  }
#endif
}


/* Check addresses before doing OSFILE */
void check_addresses()
{
#ifdef Z80_IO_RO
/*  2345 6789 ABCD EF01
    load exec leng attr
R0   R2   R3   R4   R5
255 load flag             Load file with <File$Path>
0             start end   Save file
1    -
|    -                    Read/Write info with <File$Path>
v    -
9    -
10            start end   Save file with type 
11   -                    Create empty with type
12  load flag path        Load file with path
13            path        Read info with path
14  load flag path        Load file with <path>
15            path        Read info with <path>
16  load flag             Load with no path
17   -                    Read info with no path
18   -                    Set type and stamp
19                        Create error
*/
if (Areg > 11 && Areg < 16)  adjustadd(&ctrl.start,0);
if (Areg == 0 || Areg == 10) adjustadd(&ctrl.start,&ctrl.end);
if (Areg == 255 || Areg == 12 || Areg == 14 || Areg == 16)
  if ((ctrl.exec & 255) == 0) adjustadd(&ctrl.load,0);
#endif
}


/* ==================================================================== */
/* OSFILE - Access whole files						*/
/* ==================================================================== */
void MOS_FILE(void)
{
#ifndef Z80_FILE_RO
int loadL,execL,startL,endL;
int loadH,execH,startH,endH;
#endif

HLreg=Lreg | (Hreg << 8);
addr=mem[HLreg] | (mem[HLreg+1] << 8);
io_filename(addr);			/* Copy filename to iobuffer	*/

#ifdef Z80_FILE_RO
ctrl.load=mem[HLreg+2]+(mem[HLreg+3] << 8)+(mem[HLreg+4] << 16)+(mem[HLreg+5] << 24);
ctrl.exec=mem[HLreg+6]+(mem[HLreg+7] << 8)+(mem[HLreg+8] << 16)+(mem[HLreg+9] << 24);
ctrl.start=mem[HLreg+10]+(mem[HLreg+11] << 8)+(mem[HLreg+12] << 16)+(mem[HLreg+13] << 24);
ctrl.end=mem[HLreg+14]+(mem[HLreg+15] << 8)+(mem[HLreg+16] << 16)+(mem[HLreg+17] << 24);
check_addresses();			/* Check if addresses need altering */
chk_err(tmp=_kernel_osfile(Areg,iobuffer,&ctrl));
Areg=tmp;
mem[HLreg+2]=ctrl.load;
mem[HLreg+3]=ctrl.load >> 8;
mem[HLreg+4]=ctrl.load >> 16;
mem[HLreg+5]=ctrl.load >> 24;

mem[HLreg+6]=ctrl.exec;
mem[HLreg+7]=ctrl.exec >> 8;
mem[HLreg+8]=ctrl.exec >> 16;
mem[HLreg+9]=ctrl.exec >> 24;

mem[HLreg+10]=ctrl.start;
mem[HLreg+11]=ctrl.start >> 8;
mem[HLreg+12]=ctrl.start >> 16;
mem[HLreg+13]=ctrl.start >> 24;

mem[HLreg+14]=ctrl.end;
mem[HLreg+15]=ctrl.end >> 8;
mem[HLreg+16]=ctrl.end >> 16;
mem[HLreg+17]=ctrl.end >> 24;
#endif

#ifndef Z80_FILE_RO
loadL =mem[HLreg+2] +(mem[HLreg+3]<<8);  loadH =mem[HLreg+4] +(mem[HLreg+5]<<8);
execL =mem[HLreg+6] +(mem[HLreg+7]<<8);  execH =mem[HLreg+8] +(mem[HLreg+9]<<8);
startL=mem[HLreg+10]+(mem[HLreg+11]<<8); startH=mem[HLreg+12]+(mem[HLreg+13]<<8);
endL  =mem[HLreg+14]+(mem[HLreg+15]<<8); endH  =mem[HLreg+16]+(mem[HLreg+17]<<8);

switch(Areg) {
  case 0xFF:				/* LOAD				*/
    if((execL & 0xFF) != 0) {
      z80error(254,"","LOAD address needed");  /* Generate an error	*/
      } else {
      tmp=load(iobuffer,loadL);		/* Try to load file		*/
      if (tmp == -1) {			/* Couldn't load file		*/
	z80error(214,iobuffer," not found");  /* Generate an error	*/
	} else {
	mem[HLreg+10]=tmp;		/* Return length in control block */
	mem[HLreg+11]=tmp >> 8;
	mem[HLreg+12]=0; mem[HLreg+13]=0;
	Areg=1;				/* Type=FILE			*/
	}
      }
    return;
    
  case 0x00:				/* SAVE				*/
    tmp=save(iobuffer,startL,endL,loadL,execL);
    if (tmp == -1) {			/* Couldn't save file		*/
      z80error(214,"Couldn't save",iobuffer);  /* Generate an error	*/
      } else {
      mem[HLreg+10]=tmp;		/* Return length in control block */
      mem[HLreg+11]=tmp >> 8;
      mem[HLreg+12]=0; mem[HLreg+13]=0;
      Areg=1;				/* Type=FILE			*/
      }
    return;

  case 0x05:				/* Read file info		*/
    Areg=0;				/* Unsupported just now		*/
    return;
  }
#endif
}


/* ==================================================================== */
/* OSARGS - Information on open files					*/
/* ==================================================================== */
void MOS_ARGS(void)
{           
HLreg=Lreg | (Hreg << 8);
#ifdef Z80_FILE_RO
chk_err(tmp=_kernel_osargs(Areg,Ereg,mem[HLreg]+(mem[HLreg+1] << 8)+(mem[HLreg+2] << 16)+(mem[HLreg+3] << 24)));
if (Areg == 0) Areg=tmp;
  else {
  mem[HLreg]=tmp;
  mem[HLreg+1]=tmp >> 8;
  mem[HLreg+2]=tmp >> 16;
  mem[HLreg+3]=tmp >> 24;
  Areg=0;
  }
#endif
}


/* ==================================================================== */
/* OSBGET - Read a byte from files					*/
/* ==================================================================== */
void MOS_BGET(void)
{
#ifdef Z80_FILE_RO
chk_err(Areg=_kernel_osbget(Hreg));
#endif
Freg=(Freg & 254) | ((Areg & 256) >> 8);
}


/* ==================================================================== */
/* OSBPUT - Write a byte to files					*/
/* ==================================================================== */
void MOS_BPUT(void)
{
#ifdef Z80_FILE_RO
chk_err(_kernel_osbput(Areg,Hreg));
#endif
}


/* ==================================================================== */
/* OSGBPB - Read and Write multiple bytes and information		*/
/* ==================================================================== */
void MOS_GBPB(void)
{
int flag=0;

HLreg=Lreg | (Hreg << 8);

#ifdef Z80_FILE_RO
  gbpb.dataptr =(void *)(mem[HLreg+1] | (mem[HLreg+2] << 8) | (mem[HLreg+3] << 16) | (mem[HLreg+4] << 24));
  gbpb.nbytes  =mem[HLreg+5] | (mem[HLreg+6] << 8) | (mem[HLreg+7] << 16) | (mem[HLreg+8] << 24);
  gbpb.fileptr =mem[HLreg+9] | (mem[HLreg+10] << 8) | (mem[HLreg+11] << 16) | (mem[HLreg+12] << 24);
  gbpb.wild_fld=(char *)(mem[HLreg+13] | (mem[HLreg+14] << 8) | (mem[HLreg+15] << 16) | (mem[HLreg+16] << 24));
  if (flag=(((int)gbpb.dataptr & 0xFFFF0000) == 0))
    gbpb.dataptr=(void *)((int)gbpb.dataptr + (int)&mem[0]);  /* Adjust */
  chk_err(tmp=_kernel_osgbpb(Areg, mem[HLreg], &gbpb));
  Areg=tmp;
  Freg=(Freg & 254) | ((tmp & 65536) >> 16);
/* Where is mem[HLreg+0] returned? */
  if (flag) gbpb.dataptr=(void *)((int)gbpb.dataptr - (int)&mem[0]); /* De-adjust */

  mem[HLreg+1]=(int)gbpb.dataptr;
  mem[HLreg+2]=(int)gbpb.dataptr >> 8;
  mem[HLreg+3]=(int)gbpb.dataptr >> 16;
  mem[HLreg+4]=(int)gbpb.dataptr >> 24;

  mem[HLreg+5]=gbpb.nbytes;
  mem[HLreg+6]=gbpb.nbytes >> 8;
  mem[HLreg+7]=gbpb.nbytes >> 16;
  mem[HLreg+8]=gbpb.nbytes >> 24;

  mem[HLreg+9]=gbpb.fileptr;
  mem[HLreg+10]=gbpb.fileptr >> 8;
  mem[HLreg+11]=gbpb.fileptr >> 16;
  mem[HLreg+12]=gbpb.fileptr >> 24;
#endif
}


/* ==================================================================== */
/* OSFIND - Open and close files					*/
/* ==================================================================== */
void MOS_FIND(void)
{
if (Areg != 0) {
  io_filename(Lreg | (Hreg << 8));	/* copy name to iobuffer	*/
#ifdef Z80_FILE_RO
  chk_err(tmp=_kernel_osfind(Areg,iobuffer));
#else
  tmp=0;
#endif
  Areg=tmp;
  } else {
#ifdef Z80_FILE_RO
  chk_err(tmp=_kernel_osfind(Areg,(char *)Hreg));
#endif
  }
}


/* ==================================================================== */
/* OSMISC - Miscellaneous emulator operations				*/
/* ==================================================================== */
void MOS_FF0C(void)
{
int flags;

switch(Areg) {

case 0:					/* Start Basic			*/
#ifdef Z80_FILE_RO
  if(load("<Z80Tube$Basic>",0x0100) != -1) PC=&mem[0]+0x0100;
#endif
#ifdef Z80_FILE_UNIX
  if(load("Z80$BASIC",0x0100) != -1) PC=&mem[0]+0x0100;
#endif
#ifdef Z80_FILE_DOS
/*  if (GetEnvironmentVariable("Z80$BASIC",iobuffer,254) != 0) {
    if(load(iobuffer,0x0100) != -1) PC=&mem[0]+0x0100;
    } */
#endif
#ifdef Z80_FILE_WIN
  if (GetEnvironmentVariable("Z80$BASIC",iobuffer,254) != 0) {
    if(load(iobuffer,0x0100) != -1) PC=&mem[0]+0x0100;
    }
#endif
  break;

case 1:					/* Mount disk image 		*/
  HLreg=Lreg | (Hreg << 8);		/* Point at image name		*/
  Areg = disk_mount(Breg, (char *) &mem[0] + HLreg);
  break;

case 2:					/* Dismount disk		*/
  Areg = disk_dismount(Breg);
  break;

case 3:					/* Return disk information	*/
  HLreg=Lreg | (Hreg << 8);		/* Point at space for name	*/
  if (HLreg + Creg >= 0x10000) Creg = 0x10000 - HLreg;
  Areg = disk_info(Breg, Creg, (char *) &mem[0] + HLreg, &flags);
  Creg = (int8) flags;
  break;

default:
  break;
  }
}


/* ==================================================================== */
/* OSRDINF - Read emulator system information				*/
/* ==================================================================== */
void MOS_FF0E(void)
{
HLreg=Lreg | (Hreg << 8);
tmp=(int)&mem[0];
switch (Areg & 0x7F) {
  case 0:				/* Read start of Z80 memory	*/
    break;
  case 1:				/* Read error iobuffer address	*/
    tmp+=0xFF09;              
    break;
  case 2:				/* Read escape flag address	*/
    tmp+=0xFF80;
    break;
  case 3:				/* Read error vector		*/
    tmp+=0xFFFA;
    break;
  case 4:				/* Read event vector		*/
    tmp+=0xFFFC;
    break;
  case 5:				/* Read int vector		*/
    tmp+=0xFFFE;
    break;
  case 6:				/* Read NMI address		*/
    tmp+=0x0066;
    break;
  }
if(Areg < 0x80) {
  Freg=Freg & 254;
  tmp-=(int)&mem[0];
  if(tmp < 0 || tmp > 0xFFFF) Freg++;
  Lreg=tmp;
  Hreg=tmp >> 8;
  } else {
  mem[HLreg]=tmp;
  mem[HLreg+1]=tmp >> 8;
  mem[HLreg+2]=tmp >> 16;
  mem[HLreg+3]=tmp >> 24;
  }
}


/* ==================================================================== */
/* OSWRINF - Write emulator system information				*/
/* ==================================================================== */
void MOS_FF0F(void)
{
}


/* ==================================================================== */
/* String support routines						*/
/* ==================================================================== */

int skp_spc(p)				/* Skip spaces in iobuffer	*/
int p;
{
while(iobuffer[p] == 32) p++;
return(p);
}

/* Copies CR/SPC-terminated string into iobuffer with zero terminator	*/
int io_filename(int16 ptr)
{
int l;

while(mem[ptr]==32) ptr++;
for (l=0; mem[ptr]>32 && l<250; ptr++) iobuffer[l++]=mem[ptr];
iobuffer[l]=0;
return(l);				/* Returns length		*/
}


/* ==================================================================== */
/* Error handling							*/
/* ==================================================================== */

#ifdef Z80_FILE_RO
/* Check if an error occured after calling RISC OS kernel		*/
int chk_err(int retval)
{
int l;

if (retval != -2) return FALSE;		/* No error			*/
oserr=*_kernel_last_oserror();		/* Find the error block		*/
mem[0xFF09]=0xFF;			/* RST &38 opcode		*/
mem[0xFF0A]=oserr.errnum;		/* Copy error message to iobuffer */
for(l=0; oserr.errmess[l] && l < 126; l++) mem[0xFF0B+l]=oserr.errmess[l];
mem[0xFF0B+l]=0;
mem[0xFF82]=9; mem[0xFF83]=0xFF;	/* Fault pointer		*/
/* Lreg=0x0A; Hreg=0xFF; PC=mem[0xFFFA] | (mem[0xFFFB] << 8); */
PC=&mem[0]+0xFF09;			/* 'Jump' to error		*/
return TRUE;
}
#endif


/* Generate an error within the emulator				*/
z80error(int num, char *s1, char *s2)
{
int l,m;

for(l=0; s1[l]; l++) mem[0xff0b+l]=s1[l];
for(m=0; s2[m]; m++) mem[0xff0b+(l++)]=s2[m];
mem[0xff0b+l]=0;
mem[0xff09]=0xff;
mem[0xff0a]=num;
mem[0xFF82]=0x0a;
mem[0xFF83]=0xFF;
PC=&mem[0]+0xFF09;			/* Force a jump to &FF09	*/
}


/* ==================================================================== */
/* I/O routines								*/
/* ==================================================================== */

void io_OUT(int port, int value)	/* Write a byte to output port	*/
{
printf("OUT (%04X),%d\x0a\x0d",port,value);
}

int io_IN(int port)			/* Read a byte from input port	*/
{
printf("IN (%04X)\x0a\x0d",port);
return(0);
}
