/* memory.c - Z80 CPU engine system and memory
 * Copyright (C) 1987-2003 J.G.Harston

  See the accompanying file LICENSE for terms of use.
 */

#include <stdio.h>

/* Define some known types and values					*/
#define int8 unsigned char
#define int16 int
#define TRUE 1
#define FALSE 0

/* Size of emulated I/O memory						*/
#define IOMSIZE 4096

int16 HLreg;				/* HL register			*/
int16 addr;				/* addr from mem[HLreg, HLreg+1] */
int   addr32;				/* addr from mem[HLreg0..3]	*/
char  iobuffer[256];			/* Filename, etc iobuffer	*/

/* The z80 engine CPU register set and holding variables		*/
int8 Areg, Breg, Creg, Dreg, Ereg, Hreg, Lreg, Freg;  /* Main Registers */
int8 Aalt, Balt, Calt, Dalt, Ealt, Halt, Lalt, Falt;  /* Alternate set  */
int16 XYreg[2]; int XY;
int16 SP;
int8 *PC;

int Rreg=0;
int8 Ireg=0;
int IFF1=0, IFF2=0;
int IM=0;

int S, Z, b5, H, b3, PV, N, Cy, Cy2;	/* Flag bits			*/

int quit=FALSE;				/* flag to exit the execution loop */
int quiet=FALSE;

int debug=0;				/* debug flag			*/

void (*f)(void);			/* Pointer to a function	*/
int8 byte;				/* byte read from memory	*/
int off;				/* Index offset			*/
int addr,tmp,tmp2,val;			/* Temporary variables		*/

/* ==================================================================== */
/* First eight bytes of zero page, then 256 bytes of mini-MOS		*/
/* ==================================================================== */
/* int8 mem[65665]={ 0xC3, 0x06, 0xFF, 0x00, 0x00, 0xC3, 0x00, 0xFF, */

int8 *mem;				/* Z80 memory			*/
int8 *iomem;				/* I/O processor memory		*/

int8 mem0[]={ 0xC3, 0x06, 0xFF, 0x00, 0x00, 0xC3, 0x00, 0xFF,
  0xC3, 0x08, 0xFF, 0xC3, 0x09, 0xFF, 0xED, 0x00,
  0xC9, 0xCD, 0xB3, 0xFF, 0x5A, 0x38, 0x30, 0x20, 
  0x45, 0x6D, 0x75, 0x6C, 0x61, 0x74, 0x6F, 0x72, 
  0x20, 0x54, 0x69, 0x6E, 0x79, 0x20, 0x4D, 0x4F, 
  0x53, 0x20, 0x30, 0x2E, 0x35, 0x31, 0x20, 0x28, 
  0x31, 0x33, 0x20, 0x4A, 0x75, 0x6C, 0x20, 0x31, 
  0x39, 0x39, 0x35, 0x29, 0x0D, 0x00, 0x18, 0x0E, 
  0x2A, 0x82, 0xFF, 0xE5, 0xE1, 0xCD, 0xE7, 0xFF, 
  0xCD, 0x99, 0xFF, 0xCD, 0xE7, 0xFF, 0x21, 0x38, 
  0xFF, 0xF9, 0xCD, 0xBF, 0xFF, 0x21, 0xB9, 0xFF, 
  0x22, 0x04, 0xFF, 0x3E, 0x7C, 0xCD, 0xF4, 0xFF, 
  0xCD, 0xB3, 0xFF, 0x5A, 0x38, 0x30, 0x2A, 0x00, 
  0xAF, 0x21, 0x71, 0xFF, 0xCD, 0xF1, 0xFF, 0x38, 
  0x0C, 0x21, 0x09, 0xFF, 0xCD, 0xF7, 0xFF, 0x18, 
  0xD5, 0x09, 0xFF, 0x2B, 0x20, 0xFF, 0x11, 0x45, 
  0x73, 0x63, 0x61, 0x70, 0x65, 0x00, 0xED, 0x4D, 
  0x00, 0x00, 0x09, 0xFF, 0x38, 0xFF, 0x3A, 0xBC, 
  0xFF, 0x32, 0x38, 0x00, 0x2A, 0xBD, 0xFF, 0x22, 
  0x39, 0x00, 0xC9, 0x7E, 0xB7, 0xC8, 0xCD, 0xE3, 
  0xFF, 0x23, 0x18, 0xF7, 0xE3, 0xCD, 0x93, 0xFF, 
  0xE3, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 
  0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 
  0xC9, 0xC9, 0xC9, 0xC3, 0x9C, 0xFF, 0x79, 0x18, 
  0x35, 0xC3, 0x46, 0xFF, 0xC3, 0x3C, 0xFF, 0xC3, 
  0x86, 0xFF, 0xC9, 0xC9, 0xC9, 0xAF, 0xC9, 0xC9, 
  0xAF, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xED, 0x0B, 
  0xC9, 0xED, 0x0A, 0xC9, 0xED, 0x09, 0xC9, 0xED, 
  0x08, 0xC9, 0xED, 0x07, 0xC9, 0xED, 0x06, 0xC9, 
  0xED, 0x05, 0xC9, 0xFE, 0x0D, 0x20, 0x07, 0x3E, 
  0x0A, 0xCD, 0xEE, 0xFF, 0x3E, 0x0D, 0xED, 0x04, 
  0xC9, 0xED, 0x03, 0xC9, 0xED, 0x02, 0xC9, 0xED, 
  0x01, 0xC9, 0x38, 0xFF, 0xA1, 0xFF, 0x7E, 0xFF, 
  0x00, 0x00, 0x00, 0x00, 0xC3, 0x00, 0x00 };
/* The mini-mos is moved up to &FF00...&FFFF on startup			*/

/* When debug is turned on, various info printed:
 -debug 0 - 00110000 - off
 -debug 1 - 00110001 - register dump
 -debug 2 - 00110010 - MOS calls
 -debug 4 - 00110100
 -debug H - 01001000
 -debug P - 01010000

Register dump is following format:
AF=0000 MZ5H3VNC BC=0000 DE=0000 HL=0000 IX=0000 I=00 PC=0000:00,00,00,00
AF'0000 MZ5H3VNC BC'0000 DE'0000 HL'0000 IY=0000 R=00 SP=0000:00,00,00,00
 */
 
void disp_regs()
{
printf("AF=%02X%02X ",Areg,Freg);	/* Main registers		*/
if(Freg & 128) putchar('M'); else putchar('P');
if(Freg & 64)  putchar('Z'); else putchar('.');
if(Freg & 32)  putchar('5'); else putchar('.');
if(Freg & 16)  putchar('H'); else putchar('.');
if(Freg & 8)   putchar('3'); else putchar('.');
if(Freg & 4)   putchar('V'); else putchar('.');
if(Freg & 2)   putchar('N'); else putchar('.');
if(Freg & 1)   putchar('C'); else putchar('.');
printf(" BC=%02X%02X DE=%02X%02X HL=%02X%02X",Breg,Creg,Dreg,Ereg,Hreg,Lreg);
printf(" IX=%04X I=%02X PC=%04X",XYreg[0],Ireg,PC-&mem[0]);
printf(":%02X,%02X,%02X,%02X\n",*(PC),*(PC+1),*(PC+2),*(PC+3));

printf("AF'%02X%02X ",Aalt,Falt);	/* Alternate registers		*/
if(Falt & 128) putchar('M'); else putchar('P');
if(Falt & 64)  putchar('Z'); else putchar('.');
if(Falt & 32)  putchar('5'); else putchar('.');
if(Falt & 16)  putchar('H'); else putchar('.');
if(Falt & 8)   putchar('3'); else putchar('.');
if(Falt & 4)   putchar('V'); else putchar('.');
if(Falt & 2)   putchar('N'); else putchar('.');
if(Falt & 1)   putchar('C'); else putchar('.');
printf(" BC'%02X%02X DE'%02X%02X HL'%02X%02X",Balt,Calt,Dalt,Ealt,Halt,Lalt);
printf(" IY=%04X R=%02X SP=%04X",XYreg[1],Rreg,SP);
printf(":%02X,%02X,%02X,%02X\n",mem[SP],mem[(SP+1) & 0xFFFF],mem[(SP+2) & 0xFFFF],mem[(SP+3) & 0xFFFF]);
}

